/*
 * meterBusClient.cpp
 *
 *  Created on: 08.03.2014
 *      Author: wn
 */



#include "meterBusClient.h"
#include "uart.h"
#include "time.h"
#include "debug.h"

const bool MBC_COMPILE_TIME_DEBUG = false;



MeterBusClient::MeterBusClient() : m_address(0),
    m_frameCnt(0), m_accessCnt(0), m_myFrameCnt(0), m_invalidFrameCnt(0), m_invalidChecksum(0),
    m_collisionCnt(0) {
}



void MeterBusClient::begin() {
  // setAddress(Config::getUChar(Config::METERBUSCLIENT_ADDRESS));
  setAddress(0x22);
}

void MeterBusClient::setAddress(unsigned char a) {
  // Config::setUChar(Config::METERBUSCLIENT_ADDRESS, a);
  m_address = a;
}

unsigned char MeterBusClient::getAddress() {
  return m_address;
}


unsigned char MeterBusClient::getStatus() {
  return 0;
}

unsigned char MeterBusClient::getAccessCnt() {
  return m_accessCnt;
}


/*
 * Single Character: E5h
 *
 * Short Frame:
 *   Start      10h
 *   C-Field
 *   A-Field
 *   Check-Sum
 *   Stop       16h
 *
 * Control Frame:
 *   Start      68h
 *   Length     = 3
 *   Length
 *   Start      68h
 *   C-Field
 *   A-Field
 *   CI-Field
 *   Check-Sum
 *   Stop       16h
 *
 * Long Frame:
 *   Start      68h
 *   Length
 *   Length
 *   Start      68h
 *   C-Field
 *   A-Field
 *   CI-Field
 *   User-Data
 *   Check-Sum
 *   Stop       16h
 *
 * Frames:
 *   SND_NKE
 *     Short Frame, C: 40h
 *
 *   REQ_UD2
 *     Short Frame, C: 5Bh
 */


bool isChecksumValid(MeterBusFrame frame) {
  unsigned char ctrlsum = frame.aField + frame.cField + frame.ciField;
  for (int i = 0; i < (frame.length - 3); i++) {
    ctrlsum += frame.userData[i];
  }
  return (ctrlsum == frame.checksum);
}



bool MeterBusClient::handleFrame() {
  bool res = false;
  m_frameCnt++;
  if (! isChecksumValid(m_frame)) {
    m_invalidChecksum++;
  } else if ((m_frame.aField == getAddress()) || (m_frame.aField == 0xfe)) {
    m_myFrameCnt++;
    // handle the frame

    if (m_frame.startDelimiter == 0x10) {
      if (m_frame.cField == 0x40)  {
        m_accessCnt++;
        SND_NKE();
        res = true;
      } else if (m_frame.cField == 0x5b) {
        m_accessCnt++;
        REQ_UD2();
        res = true;
      }
    } else if (m_frame.startDelimiter == 0x68) {
      if (m_frame.cField == 0x53) {
        m_accessCnt++;
        SND_UD();
        res = true;
      }
    }
  }
  return res;
}


typedef enum {
  STATE_START, STATE_IDLE, STATE_SHORT_FRAME, STATE_LONG_CTRL_FRAME, STATE_HANDLE, STATE_INVALID, STATE_DONE,
  STATE_DELAYED_RESPONSE, STATE_DELAY, STATE_RESPONSE, STATE_RECEIVE_ECHO
} e_meterBusClientState;

typedef enum {
  SUBSTATE_LENGTH, SUBSTATE_LENGTH_REPEAT, SUBSTATE_START_REPEAT, SUBSTATE_C_FIELD, SUBSTATE_A_FIELD, SUBSTATE_CI_FIELD,
  SUBSTATE_USERDATA, SUBSTATE_CHECKSUM, SUBSTATE_STOP
} e_meterBusClientSubState;


void MeterBusClient::exec() {
  static e_meterBusClientState state = STATE_IDLE;
  static e_meterBusClientSubState subState = SUBSTATE_LENGTH;
  static unsigned char userDataCnt = 0;
  static unsigned int echoReceiveCnt = 0;
  static unsigned long receivedDoneTimestamp;


  int chi = uartRead();
  char ch = (char) chi;
  if (chi != -1) {
    debugWrite(0x04);
    debugWrite(ch);
    debugWrite(state);
    debugWrite(subState);
  }

  switch (state) {
  case STATE_START:
    m_frame.startDelimiter = 0;
    m_frame.length = 0;
    m_frame.cField = 0;
    m_frame.aField = 0;
    m_frame.ciField = 0;
    m_frame.valid = false;
    state = STATE_IDLE;
    break;

  case STATE_IDLE:
    if (chi != -1) {
      if (ch == 0x10) {
        m_frame.startDelimiter = 0x10;
        state = STATE_SHORT_FRAME;
        subState = SUBSTATE_C_FIELD;
      } else if (ch == 0x68)  {
        m_frame.startDelimiter = 0x68;
        state = STATE_LONG_CTRL_FRAME;
        subState = SUBSTATE_LENGTH;
      } else {
        state = STATE_INVALID;
      }
    }
    break;

  case STATE_SHORT_FRAME:
    if (chi != -1) {
      if (subState == SUBSTATE_C_FIELD) {
        m_frame.cField = ch;
        subState = SUBSTATE_A_FIELD;
      } else if (subState == SUBSTATE_A_FIELD) {
        m_frame.aField = ch;
        subState  = SUBSTATE_CHECKSUM;
      } else if (subState == SUBSTATE_CHECKSUM) {
        m_frame.checksum = ch;
        subState  = SUBSTATE_STOP;
      } else if (subState == SUBSTATE_STOP) {
        if (ch == 0x16) {
          m_frame.valid = true;
          state = STATE_HANDLE;
        } else {
          state = STATE_INVALID;
        }
      }
    }
    break;

  case STATE_LONG_CTRL_FRAME:
    if (chi != -1) {
      if (subState == SUBSTATE_LENGTH) {
        m_frame.length = ch;
        if ((m_frame.length < 3) || (m_frame.length > 252)) {
          state = STATE_INVALID;
        } else {
          subState = SUBSTATE_LENGTH_REPEAT;
        }
      } else if (subState == SUBSTATE_LENGTH_REPEAT) {
        if (ch == m_frame.length) {
          subState = SUBSTATE_START_REPEAT;
        } else {
          state = STATE_INVALID;
        }
      } else if (subState == SUBSTATE_START_REPEAT) {
        if (ch == 0x68) {
          subState = SUBSTATE_C_FIELD;
        } else {
          state = STATE_INVALID;
        }
      } else if (subState == SUBSTATE_C_FIELD) {
        m_frame.cField = ch;
        subState = SUBSTATE_A_FIELD;
      } else if (subState == SUBSTATE_A_FIELD) {
        m_frame.aField = ch;
        subState = SUBSTATE_CI_FIELD;
      } else if (subState == SUBSTATE_CI_FIELD) {
        m_frame.ciField = ch;
        if (m_frame.length == 3) {
          subState = SUBSTATE_CHECKSUM;
        } else {
          subState = SUBSTATE_USERDATA;
          userDataCnt = 0;
        }
      } else if (subState == SUBSTATE_USERDATA) {
        m_frame.userData[userDataCnt] = ch;
        userDataCnt++;
        if (userDataCnt >= (m_frame.length - 3)) {
          subState = SUBSTATE_CHECKSUM;
        }
      } else if (subState == SUBSTATE_CHECKSUM) {
        m_frame.checksum = ch;
        subState = SUBSTATE_STOP;
      } else if (subState == SUBSTATE_STOP) {
        if (ch == 0x16) {
          m_frame.valid = true;
          state = STATE_HANDLE;
        } else {
          state = STATE_INVALID;
        }
      }
    }
    break;

  case STATE_INVALID:
    m_invalidFrameCnt++;
    state = STATE_START;
    break;

  case STATE_HANDLE:
    state = STATE_DONE;
    if (m_frame.valid) {
      receivedDoneTimestamp = millis();
      if (handleFrame()) {
        state = STATE_DELAY;
      }
    }
    break;

  case STATE_DELAY:
    if ((receivedDoneTimestamp + RESPONSE_DELAY) < millis()) {
      state = STATE_DELAYED_RESPONSE;
    }
    break;

  case STATE_DELAYED_RESPONSE:
    for (unsigned int i = 0; i < m_sendBufferLen; i++) {
      uartWrite(m_sendBuffer[i]);
    }
    echoReceiveCnt = 0;
    state = STATE_RECEIVE_ECHO;
    break;

  case STATE_RECEIVE_ECHO:
    if (chi != -1) {
      if (ch == m_sendBuffer[echoReceiveCnt]) {
        echoReceiveCnt++;
        if (echoReceiveCnt == m_sendBufferLen) {
          state = STATE_DONE;
        }
      } else {
        m_collisionCnt++;
        state = STATE_DONE;
      }
    }
    break;

  case STATE_DONE:
    // may be another useful message
    state = STATE_START;
    break;
  }
}


